plot.measurements <- function(drone.df, u = NULL, dims = c(1, 2)){
    ## Extracting columns.
    id <- drone.df$id
    y <- drone.df[, names(drone.df) != "id"]
    ## Converting y to a matrix.
    if (is.data.frame(y)){
        y <- as.matrix(y)
    }
    ## Number of mantas.
    n.mantas <- length(unique(id))
    ## Choosing a colour palette for the plot.
    cols <- hcl.colors(n.mantas, palette = "viridis")
    ## Mapping colours to ids.
    col.id.map <- data.frame(cols, unique(id))
    point.cols <- character(nrow(y))
    for (i in 1:nrow(y)){
        point.cols[i] <- cols[which(id[i] == sort(unique(id)))]
    }
    ## Plotting photo measurements colour-coded by individual.
    plot(y[, dims], col = point.cols)
    ## Adding real measurement values.
    if (!is.null(u)){
        points(u[, dims], col = cols, pch = 16)
    }
}

construct.cov <- function(rho, sd){
    ## Turning arguments into matrices if provided as vectors for a single group.
    if (is.null(dim(rho))){
        rho <- matrix(rho, nrow = 1)
        sd <- matrix(sd, nrow = 1)
    }
    ## Number of dimensions.
    n.dims <- ncol(sd)
    ## Number of groups.
    n.groups <- nrow(sd)
    ## Setting up matrix arrays.
    vcov.arr <- array(0, dim = c(n.dims, n.dims, n.groups))
    cor.arr <- array(0, dim = c(n.dims, n.dims, n.groups))
    for (g in 1:n.groups){
        cor.arr[, , g] <- diag(n.dims)
    }
    for (g in 1:n.groups){
        k <- 1
        for (i in 1:(n.dims - 1)){
            for (j in (i + 1):n.dims){
                cor.arr[i, j, g] <- cor.arr[j, i, g] <- rho[g, k]
                k <- k + 1
            }
        }
        for (i in 1:n.dims){
            for (j in 1:n.dims){
                vcov.arr[i, j, g] <- cor.arr[i, j, g]*sd[g, i]*sd[g, j]
            }
        }
    }
    vcov.arr
}

## Helper function to organise data.
organise.data <- function(drone.df = NULL, individual.df = NULL, error.dist, n.pred.vals = 50, orig.data = NULL){
    n.mantas <- nrow(individual.df)
    if (!all(sort(individual.df$id) == 1:n.mantas)){
        stop("IDs must be sequential from 1 to the total number of individuals.")
    }
    if (!all(sort(unique(drone.df$id)) == sort(unique(individual.df$id)))){
        stop("All individuals in drone.df must have an entry in individual.df.")
    }
    if (is.null(orig.data)){
        n.dims <- ncol(drone.df) - 1
        n.groups <- length(unique(individual.df$group))
    } else {
        n.dims <- orig.data$n_dims
        n.groups <- orig.data$n_groups
    }
    if (is.null(drone.df)){
        no.y <- 1
        y <- as.matrix(0, nrow = 1, ncol = n.dims)
        y.id <- 0
    } else {
        y <- as.matrix(drone.df[, colnames(drone.df) != "id"])
        y.id <- drone.df$id
        no.y <- 0
    }
    group <- individual.df$group
    if (ncol(individual.df) == n.dims + 2){
        u.known <- individual.df[, -(1:2)]
    } else {
        u.known <- matrix(NA, nrow = n.mantas, ncol = n.dims)
    }
    u.is.known <- !is.na(u.known)
    if (error.dist == "mvn"){
        error.mvn <- 1
    } else if (error.dist == "mvln"){
        error.mvn <- 0
    } else {
        stop("The error.dist argument must either be \'mvn\' or \'mvln\'.")
    }
    if (n.pred.vals > 0){
        y.lims <- apply(y, 2, range)
        x.pred.vals <- matrix(0, nrow = n.dims, ncol = n.pred.vals)
        for (i in 1:n.dims){
            x.pred.vals[i, ] <- seq(y.lims[1, i], y.lims[2, i], length.out = n.pred.vals)
        }
        do.preds <- 1
    } else {
        x.pred.vals <- as.matrix(0)
        do.preds <- 0
    } 
    logit.pop.rho <- matrix(0, nrow = n.groups, ncol = sum(1:(n.dims - 1)))
    log.pop.sd <- matrix(rep(log(apply(y, 2, sd)), each = n.groups), nrow = n.groups, ncol = n.dims)
    identity.pop.mu <- matrix(rep(apply(y, 2, mean), each = n.groups), nrow = n.groups, ncol = n.dims)
    logit.me.rho <- numeric(sum(1:(n.dims - 1)))
    log.me.sd <- numeric(n.dims)
    u.known <- c(t(u.known))
    u.unknown <- matrix(rep(apply(y, 2, mean), each = n.mantas), nrow = n.mantas, ncol = n.dims)
    u.unknown <- c(t(u.unknown))
    u.unknown <- u.unknown[is.na(u.known)]
    u.known <- u.known[!is.na(u.known)]
    tmb.data <- list(n_mantas = n.mantas,
                     n_dims = n.dims,
                     n_groups = n.groups,
                     y = y,
                     y_id = y.id,
                     group = group,
                     u_is_known = u.is.known,
                     u_known = u.known,
                     error_mvn = error.mvn,
                     n_pred_vals = n.pred.vals,
                     x_pred_vals = x.pred.vals,
                     no_y = no.y,
                     do_preds = do.preds)
    tmb.pars <- list(u_unknown = u.unknown,
                     logit_pop_rho = logit.pop.rho,
                     log_pop_sd = log.pop.sd,
                     identity_pop_mu = identity.pop.mu,
                     logit_me_rho = logit.me.rho,
                     log_me_sd = log.me.sd)
    list(data = tmb.data, pars = tmb.pars)
}

fit.hmv <- function(drone.df, individual.df = NULL, error.dist = "mvn"){
    od <- organise.data(drone.df, individual.df, error.dist)
    compile("model.cpp")
    dyn.load(dynlib("model"))
    ## Creating TMB model object.
    model.obj <- MakeADFun(data = od$data,
                           parameters = od$pars,
                           random = "u_unknown", DLL = "model")
    ## Fitting model.
    fit <- nlminb(model.obj$par, model.obj$fn, model.obj$gr,
                  control = list(eval.max = 500))
    ## Manually calculating Hessian.
    hess <- optimHess(fit$par, model.obj$fn, model.obj$gr)
    ## Generating SD report.
    fit.sdreport <- sdreport(model.obj)
    ## For prediction:
    ## (1) Create new model.obj with new u values to predict.
    ## (2) With new model.obj, run sdreport with fixed pars and hess set at values from fitted model.
    ## (3) Then just get new predicted u from summary(sdreport).
    par.summary <- summary(fit.sdreport, select = "report")
    u.summary <- par.summary[rownames(par.summary) == "u", ]
    par.summary <- par.summary[rownames(par.summary) != "u", ]
    ## y.pred.vals <- matrix(par.summary[rownames(par.summary) == "y_pred_vals", 1], nrow = 1000, byrow = TRUE)
    ## y.pred.ses <- matrix(par.summary[rownames(par.summary) == "y_pred_vals", 2], nrow = 1000, byrow = TRUE)
    ## out <- list(par.summary = par.summary, u.summary = u.summary,
    ##             y.pred.vals = y.pred.vals, y.pred.ses = y.pred.ses, drone.df = drone.df,
    ##             tmb.data = tmb.data, tmb.pars = tmb.pars, pars = fit$par,
    ##             tmb.hess = hess, loglik = -fit$objective)
    par.summary <- summary(fit.sdreport, select = "report")
    ps.names <- rownames(par.summary)
    u.ests <- matrix(par.summary[ps.names == "u", 1], ncol = od$data$n_dims)
    u.ses <- matrix(par.summary[ps.names == "u", 2], ncol = od$data$n_dims)
    pop.rho.est <- matrix(par.summary[ps.names == "pop_rho", 1], nrow = od$data$n_groups)
    pop.sd.est <- matrix(par.summary[ps.names == "pop_sd", 1], nrow = od$data$n_groups)
    pop.mu.est <- matrix(par.summary[ps.names == "pop_mu", 1], nrow = od$data$n_groups)
    pop.rho.se <- matrix(par.summary[ps.names == "pop_rho", 2], nrow = od$data$n_groups)
    pop.sd.se <- matrix(par.summary[ps.names == "pop_sd", 2], nrow = od$data$n_groups)
    pop.mu.se <- matrix(par.summary[ps.names == "pop_mu", 2], nrow = od$data$n_groups)
    me.rho.est <- par.summary[ps.names == "me_rho", 1]
    me.sd.est <- par.summary[ps.names == "me_sd", 1]
    me.rho.se <- par.summary[ps.names == "me_rho", 2]
    me.sd.se <- par.summary[ps.names == "me_sd", 2]
    y.preds <- array(par.summary[ps.names == "y_pred_vals", 1],
                     dim = c(2*sum(1:(od$data$n_dims - 1)),
                             od$data$n_pred_vals,
                             od$data$n_groups))
    y.ses <- array(par.summary[ps.names == "y_pred_vals", 2],
                     dim = c(2*sum(1:(od$data$n_dims - 1)),
                             od$data$n_pred_vals,
                             od$data$n_groups))
    ests <- list(pop.rho = pop.rho.est, pop.sd = pop.sd.est,
                 pop.mu = pop.mu.est, me.rho = me.rho.est, me.sd = me.sd.est)
    ses <- list(pop.rho = pop.rho.se, pop.sd = pop.sd.se,
                 pop.mu = pop.mu.se, me.rho = me.rho.se, me.sd = me.sd.se)
    out <- list(ests = ests, ses = ses, u.ests = u.ests, u.ses = u.ses,
                y.preds = y.preds, y.ses = y.ses, data = od$data, drone.df = drone.df,
                individual.df = individual.df, error.dist = error.dist, pars = fit$par,
                hess = hess)
    class(out) <- "hmv"
    out
}

predict.hmv <- function(fit, new.drone.df = NULL, new.individual.df = NULL){
    od <- organise.data(new.drone.df, new.individual.df, fit$error.dist, 0, fit$data)
    pred.model.obj <- MakeADFun(data = od$data,
                                parameters = od$pars,
                                random = "u_unknown", DLL = "model")
    pred.sdreport <- sdreport(pred.model.obj, fit$pars, fit$hess)
    pred.par.summary <- summary(pred.sdreport, select = "report")
    pred.u.summary <- pred.par.summary[rownames(pred.par.summary) == "u", ]
    pred.u.ests <- matrix(pred.u.summary[, 1], ncol = fit$data$n_dims)
    pred.u.ses <- matrix(pred.u.summary[, 2], ncol = fit$data$n_dims)
    list(u.ests = pred.u.ests, u.ses = pred.u.ses)
}

plot.hmv <- function(object, group = 1, dims = c(1, 2)){
    ## Figuring out which predictions and pairwise indices to use.
    k <- 1
    l <- 1
    for (i in 1:(object$data$n_dims - 1)){
        for (j in (i + 1):object$data$n_dims){
            if (isTRUE(all.equal(dims, c(i, j)))){
                pred.index <- k + 1
                pairwise.index <- l
            } else if (isTRUE(all.equal(dims, c(j, i)))){
                pred.index <- k
                pairwise.index <- l
            }
            l <- l + 1
            k <- k + 2
        }
    }
    ## Plotting everything.
    plot.measurements(object$drone.df[object$drone.df$id %in%
                                      object$individual.df$id[object$individual.df$group == group], ],
                      u = object$u.ests[object$individual.df$group == group, ],
                      dims = dims)
    ## Extracting relevant population parameters.
    pop.rho.est <- object$ests$pop.rho[group, ]
    pop.sd.est <- object$ests$pop.sd[group, ]
    pop.mu.est <- object$ests$pop.mu[group, ]
    ## Calculating prediction intercept and slope.
    pred.int <- pop.mu.est[dims[2]] - pop.rho.est[pairwise.index]*pop.sd.est[dims[2]]/pop.sd.est[dims[1]]*pop.mu.est[dims[1]]
    pred.slope <- pop.rho.est[pairwise.index]*pop.sd.est[dims[2]]/pop.sd.est[dims[1]]
    ## Calculating upper and lower CI bounds.
    pred.lower <- object$y.preds[pred.index, , group] - qnorm(0.975)*object$y.ses[pred.index, , group]
    pred.upper <- object$y.preds[pred.index, , group] + qnorm(0.975)*object$y.ses[pred.index, , group]
    lines(object$data$x_pred_vals[dims[1], ], pred.lower, lty = "dotted")
    lines(object$data$x_pred_vals[dims[1], ], pred.upper, lty = "dotted")
    lines(object$data$x_pred_vals[dims[1], ], object$y.preds[pred.index, , group])
}

plot.me <- function(fit, dims, n.xy = 200, n.cols = 100, ...){
    me.rho <- fit$ests$me.rho
    me.sd <- fit$ests$me.sd
    me.cov <- construct.cov(me.rho, me.sd)[, , 1]
    xlim <- c(-3, 3)*me.sd[dims[1]]
    ylim <- c(-3, 3)*me.sd[dims[2]]
    x <- seq(xlim[1], xlim[2], length.out = n.xy)
    y <- seq(ylim[1], ylim[2], length.out = n.xy)
    dme <- function(x, y) dmvnorm(cbind(x, y), sigma = me.cov[dims, dims])
    z <- outer(x, y, dme)
    par(xaxs = "i", yaxs = "i")
    plot.new()
    plot.window(xlim = xlim, ylim = ylim)
    axis(1)
    axis(2)
    title(...)
    box()
    image(x, y, z, xlim = xlim, ylim = ylim, col = viridis(n.cols), add = TRUE)
    abline(h = 0 , v = 0)
}

coef.calc <- function(fit, y.dim, x.dim, group){
    n.x <- length(x.dim)
    pop.mu <- fit$ests$pop.mu[group, ]
    pop.rho <- fit$ests$pop.rho[group, ]
    pop.sd <- fit$ests$pop.sd[group, ]
    pop.cov <- construct.cov(pop.rho, pop.sd)[, , 1]
    sub.cov.y <- pop.cov[y.dim, x.dim, drop = FALSE]
    sub.cov.x <- pop.cov[x.dim, x.dim, drop = FALSE]
    betas <- numeric(n.x)
    betas[2:(n.x + 1)] <- c(sub.cov.y %*% solve(sub.cov.x))
    betas[1] <- pop.mu[y.dim] - sum(betas[2:(n.x + 1)]*pop.mu[x.dim])
    names(betas) <- c("b0", paste0("b", x.dim))
    betas
}

summary.hmv <- function(object, par, ...){
    if (par %in% c("pop.mu", "pop.sd", "pop.rho")){
        out <- vector(mode = "list", length = 3)
        names(out) <- c("Estimate", "Std. Error", "T-stat")
        out[1] <- object$est[par]
        out[2] <- object$ses[par]
        out[[3]] <- out[[1]]/out[[2]]
        rownames(out[[1]]) <- rownames(out[[2]]) <- rownames(out[[3]]) <- paste0("Group", 1:nrow(out[[1]]))
        if (par == "pop.rho"){
            n.dims <- ncol(object$ests$pop.mu)
            cn <- apply(combn(1:n.dims, m = 2), 2, function(x) paste("Dim", x[1], "vDim", x[2], sep = ""))
            colnames(out[[1]]) <- colnames(out[[2]]) <- colnames(out[[3]]) <- cn
        } else {
            colnames(out[[1]]) <- colnames(out[[2]]) <- colnames(out[[3]]) <- paste0("Dim", 1:ncol(out[[1]]))
        }
    } else if (par == "me.sd"){
        out <- matrix(0, nrow = length(object$est[[par]]), ncol = 3)
        colnames(out) <- c("Estimate", "Std. Error", "T-stat")
        rownames(out) <- paste0("Dim", 1:nrow(out))
        out[, 1] <- object$est[[par]]
        out[, 2] <- object$ses[[par]]
        out[, 3] <- out[, 1]/out[, 2]
    }
    out
}
